/**************************************************************************

   File:                DivergenceSkeletonMaker.h

   Author(s):           Pavel Dimitrov

   Created:             27 Jun 2002

   Last Revision:       $Date: 2003/06/05 04:38:50 $

   Description: 

   $Revision: 1.6 $

   $Log: DivergenceSkeletonMaker.h,v $
   Revision 1.6  2003/06/05 04:38:50  pdimit
   Before making the switch to the new ordering.
   The new ordering will be
   DivPt p1,p2;

   p1 <= p2    <==>    (p1.dist < p2.dist AND abs(p1.dist - p2.dist) < sigma)
                    OR (p1.div < p2.div)

   where sigma is the lattice spacing.

   Revision 1.5  2002/07/25 20:50:47  pdimit
   Making release 0.1

   Revision 1.4  2002/07/23 21:02:51  pdimit
   The branch segmentation of the thinned DivMap is improved --
   the two pixel branches are now included. Still, the degenerate case
   of the square of junction-points is not taken care of.

   Also, a DiscreteDivergenceSkeleton is created which still does not know
   of left and right for the branches, i.e. the contour has not been cut.

   Revision 1.3  2002/07/01 07:03:09  pdimit
   Now the thinned discrete divergence array is disected properly
   into branches; that is, the algorithm seems to be sound.

   Revision 1.2  2002/06/30 05:58:43  pdimit
   Just implemented the colouring of the thinned DivergenceMap.
   Also, fixed a bug in the way of counting the number of intersection
   in the 8-nbhd of a pt.

   Revision 1.1  2002/06/30 01:22:30  pdimit
   Added DivergenceSkeletonMaker and a test for it. Incomplete functionality
   and the testing is mostly in DivergenceSkeletonMaker.cpp. Only the
   thinning of the DivergenceMap is implemented.


	
   Copyright (c) 2002 by Pavel Dimitrov, Centre for Intelligent Machines,
   McGill University, Montreal, QC.  Please see the copyright notice
   included in this distribution for full details.

 **************************************************************************/

#ifndef DIVERGENCESKELETONMAKER_H
#  define DIVERGENCESKELETONMAKER_H

#include "DivergenceMap.h"
#include "DiscreteDivergenceSkeleton.h"

#define DEBUG_DIVERGENCE_SKELETON_MAKER

namespace sg {

  const char FOREGROUND_COL       =0;
  const char BACKGROUND_COL       =1;
  const char SKELETON_COL         =2;
  const char SK_POINT_COL         =3;
  const char END_POINT_COL        =4;
  const char BRANCHING_POINT_COL  =5;


  struct DivPt {
    
    double sigma; // lattice spacing
    double thresh; // thinning threshold

    Point p;
    double val; // divergence value
    double dist; // closest distance
    
    char col; // for the thinning and skeleton
    
    bool visited; // for the skeleton creation (branches)
    
    int x, y; // array position
    
    DivPt():visited(false), x(0),y(0) { 
      dist = 0;
      col=FOREGROUND_COL; 

      sigma = 1;
    }
    
    DivPt(Point pt, double div, double sig, double thr):visited(false),
							x(0),y(0){

      sigma = sig;
      thresh = thr;
      
      dist=0;
      p = pt;
      val = div;
      col=FOREGROUND_COL;
    }
    
    bool operator<(const DivPt &dp) const{
      return val<dp.val; // since max-heap

//       if(fabs(dist - dp.dist) < 0.5*sigma)
// 	return val<dp.val; 
//       else
// 	return dist<dp.dist; // since max-heap
 
//       if((val < thresh) || (dp.val < thresh)){
	
// 	if(abs(x-dp.x) + abs(y-dp.y) <= 8)
// 	  return val<dp.val; 
// 	else
// 	  return dist<dp.dist; // since max-heap
//       }
//       else
// 	return dist<dp.dist; // since max-heap

      
//      return dist<dp.dist; // since max-heap

    }
  };
  
  ////
  //  typedef std::vector<DivPt> DivPtVector;
  
  class DivPtVector : public std::vector<DivPt> {
  public:
    DiscreteDivergenceSkeletonNode *n1, *n2;

    DivPtVector () { n1=0; n2=0; }

  };

  
  ////
  class DivArr {
    
    int xSize, ySize;
    DivPt *buf;
    
    DivPt deflt;
        
  public:
    
    DivArr(int xs, int ys){
      xSize = xs;
      ySize = ys;
      
      buf = new DivPt[xSize*ySize];
      
      deflt.val = 10000;
      deflt.col = BACKGROUND_COL;
    }
    
    DivArr(std::vector<DivPt> pts, int xs){

      xSize = xs;
      ySize = pts.size()/xs;
      
      buf = new DivPt[pts.size()];
      deflt.col = BACKGROUND_COL;
      
      std::vector<DivPt>::iterator I;
      int i;
      for(i=0, I=pts.begin(); I!=pts.end(); I++, i++){
	buf[i] = *I;
	buf[i].x = i % xs;
	buf[i].y = i / xs;
      }
      
      deflt.val = 10000;
    }
    
    ~DivArr() { delete buf; }
    
    DivPt &operator()(int x, int y){
      if(x>=0 && x<xSize && y>=0 && y<ySize)
	return buf[y*xSize + x];    
      else{
	//std::cerr << "BTTD ";
	return deflt;
      }
    }
    
    int getXSize() const { return xSize; }
    int getYSize() const { return ySize; }
    
  };
  
  ////////////////////////////////////////////////
  ////////////////////////////////////////////////
  class DivergenceSkeletonMaker{

    DivergenceMap *dm;
    DivArr *da;

  public:
    DivergenceSkeletonMaker(const DivergenceMap &dm){
      this->dm = dm.clone();
    }
    DivergenceSkeletonMaker(const DivergenceMap *dm){
      this->dm = dm->clone();
    }

    DivArr&DivergenceSkeletonMaker::getThinDivArr(double step, double thresh);

    DiscreteDivergenceSkeleton *getDiscreteDivergenceSkeleton(double step, 
							      double thresh);

    void make_skeleton(double step=1.0, double thresh=-4);
  };

}

#endif  /* DIVERGENCESKELETONMAKER_H */
